@@ -1,11 +1,10 @@
 from datetime import datetime
-from enum import Enum
-from typing import Optional, Tuple
+from typing import Dict, List, Optional, Tuple
 
 from pydantic import Field, validator
 
 from .account import Account
-from .base import BaseModel
+from .base import BaseModel, IntEnum, StrEnum
 from .experiment_spec import DataMount, EnvVar, ImageSource, Priority, TaskSpec
 
 __all__ = [
@@ -25,7 +24,7 @@ __all__ = [
 ]
 
 
-class CurrentJobStatus(str, Enum):
+class CurrentJobStatus(StrEnum):
     """
     The status of a job.
     """
@@ -41,7 +40,7 @@ class CurrentJobStatus(str, Enum):
     preempted = "preempted"
 
 
-class CanceledCode(int, Enum):
+class CanceledCode(IntEnum):
     not_set = 0
     system_preemption = 1
     user_preemption = 2
@@ -125,7 +124,7 @@ class JobExecution(BaseModel):
     workspace: Optional[str] = None
 
 
-class JobKind(str, Enum):
+class JobKind(StrEnum):
     """
     The kind of job.
     """
@@ -141,9 +140,11 @@ class Session(BaseModel):
     image: Optional[ImageSource] = None
     save_image: bool = False
     ports: Optional[Tuple[int, ...]] = None
+    ports_v2: Optional[Tuple[Tuple[int, int], ...]] = None
     priority: Optional[Priority] = None
     work_dir: Optional[str] = None
     identity: Optional[str] = None
+    constraints: Optional[Dict[str, List[str]]] = None
 
 
 class Job(BaseModel):
@@ -166,6 +167,8 @@ class Job(BaseModel):
     requests: Optional[JobRequests] = None
     limits: Optional[JobLimits] = None
     session: Optional[Session] = None
+    host_networking: bool = False
+    port_mappings: Optional[Dict[str, int]] = None
 
     @property
     def display_name(self) -> str:
@@ -182,6 +185,13 @@ class Job(BaseModel):
         """
         return self.status.current == CurrentJobStatus.finalized
 
+    @property
+    def was_preempted(self) -> bool:
+        return self.status.canceled is not None and self.status.canceled_code in {
+            CanceledCode.system_preemption,
+            CanceledCode.user_preemption,
+        }
+
     def check(self):
         """
         :raises JobFailedError: If the job failed or was canceled.
@@ -190,12 +200,13 @@ class Job(BaseModel):
 
         if self.status.exit_code is not None and self.status.exit_code > 0:
             raise JobFailedError(
-                f"Job '{self.id}' exited with non-zero exit code ({self.status.exit_code})"
+                f"Job '{self.id}' exited with non-zero exit code ({self.status.exit_code})",
+                job=self,
             )
         elif self.status.canceled is not None:
-            raise JobFailedError(f"Job '{self.id}' was canceled")
+            raise JobFailedError(f"Job '{self.id}' was canceled", job=self)
         elif self.status.failed is not None:
-            raise JobFailedError(f"Job '{self.id}' failed")
+            raise JobFailedError(f"Job '{self.id}' failed", job=self)
 
 
 class Jobs(BaseModel):
@@ -211,7 +222,8 @@ class JobStatusUpdate(BaseModel):
     failed: Optional[bool] = None
     finalized: Optional[bool] = None
     canceled: Optional[bool] = None
-    canceled_for: Optional[CanceledCode] = None
+    canceled_for: Optional[str] = None
+    canceled_code: Optional[CanceledCode] = None
     idle: Optional[bool] = None
     message: Optional[str] = None
 
