@@ -1,8 +1,18 @@
 import io
 import os
+import urllib.parse
 from datetime import datetime
 from pathlib import Path
-from typing import TYPE_CHECKING, Deque, Dict, Generator, Optional, Tuple, Union
+from typing import (
+    TYPE_CHECKING,
+    ClassVar,
+    Dict,
+    Generator,
+    List,
+    Optional,
+    Tuple,
+    Union,
+)
 
 from ..aliases import PathOrStr
 from ..data_model import *
@@ -15,6 +25,9 @@ if TYPE_CHECKING:
     from rich.progress import Progress, TaskID
 
 
+is_canceled = None
+
+
 class DatasetClient(ServiceClient):
     """
     Accessed via :data:`Beaker.dataset <beaker.Beaker.dataset>`.
@@ -27,7 +40,12 @@ class DatasetClient(ServiceClient):
     HEADER_LAST_MODIFIED = "Last-Modified"
     HEADER_CONTENT_LENGTH = "Content-Length"
 
-    REQUEST_SIZE_LIMIT = 32 * 1024 * 1024
+    REQUEST_SIZE_LIMIT: ClassVar[int] = 32 * 1024 * 1024
+
+    DOWNLOAD_CHUNK_SIZE: ClassVar[int] = 10 * 1024
+    """
+    The default buffer size for downloads.
+    """
 
     def get(self, dataset: str) -> Dataset:
         """
@@ -193,16 +211,30 @@ class DatasetClient(ServiceClient):
         max_workers: Optional[int] = None,
         quiet: bool = False,
         validate_checksum: bool = True,
+        chunk_size: Optional[int] = None,
+        multiprocessing: bool = False,
     ):
         """
         Download a dataset.
 
+        .. note::
+            This method uses a :class:`~concurrent.futures.ThreadPoolExecutor` by default to
+            download files concurrently. This is much faster than downloading files sequentially,
+            however you may find that it's still pretty slow compared to the Beaker CLI when
+            the dataset has many files. Unfortunately this is an inherent limitation of Python's GIL.
+
+            To get around this you could try setting `multiprocessing=True`, which will
+            force using a :class:`~concurrent.futures.ProcessPoolExecutor` instead.
+
         :param dataset: The dataset ID, name, or object.
         :param target: The target path to fetched data. Defaults to ``Path(.)``.
         :param max_workers: The maximum number of thread pool workers to use to download files concurrently.
         :param force: If ``True``, existing local files will be overwritten.
         :param quiet: If ``True``, progress won't be displayed.
         :param validate_checksum: If ``True``, the checksum of every file downloaded will be verified.
+        :param chunk_size: The size of the buffer (in bytes) to use while downloading each file.
+            Defaults to :data:`DOWNLOAD_CHUNK_SIZE`.
+        :param multiprocessing: Use multiprocessing instead of threading.
 
         :raises DatasetNotFound: If the dataset can't be found.
         :raises DatasetReadError: If the :data:`~beaker.data_model.dataset.Dataset.storage` hasn't been set.
@@ -221,63 +253,68 @@ class DatasetClient(ServiceClient):
         if dataset.storage is None:
             raise DatasetReadError(dataset.id)
 
-        storage_info = DatasetStorageInfo.from_json(
+        dataset_info = DatasetInfo.from_json(
             self.request(
-                f"datasets/{dataset.storage.id}",
-                method="GET",
-                token=dataset.storage.token,
-                base_url=dataset.storage.address,
+                f"datasets/{dataset.id}/files",
+                exceptions_for_status={404: DatasetNotFound(self._not_found_err_msg(dataset.id))},
             ).json()
         )
+        total_bytes_to_download: int = dataset_info.size.bytes
+        total_downloaded: int = 0
 
         target = Path(target or Path("."))
         target.mkdir(exist_ok=True, parents=True)
 
-        total_bytes_to_download: Optional[int] = None
-        total_downloaded: int = 0
-        from ..progress import (
-            get_sized_dataset_fetch_progress,
-            get_unsized_dataset_fetch_progress,
-        )
+        from ..progress import get_sized_dataset_fetch_progress
 
-        if storage_info.size is not None and storage_info.size.final:
-            total_bytes_to_download = storage_info.size.bytes
-            progress = get_sized_dataset_fetch_progress(quiet)
-        else:
-            progress = get_unsized_dataset_fetch_progress(quiet)
+        progress = get_sized_dataset_fetch_progress(quiet)
 
         with progress:
             bytes_task = progress.add_task("Downloading dataset")
-            if total_bytes_to_download is not None:
-                progress.update(bytes_task, total=total_bytes_to_download)
+            progress.update(bytes_task, total=total_bytes_to_download)
 
             import concurrent.futures
+            import multiprocessing as mp
+            import threading
 
-            with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
+            pool_type = (
+                concurrent.futures.ProcessPoolExecutor
+                if multiprocessing
+                else concurrent.futures.ThreadPoolExecutor
+            )
+            with pool_type(max_workers=max_workers) as executor:
+                global is_canceled
+                is_canceled = mp.Event() if multiprocessing else threading.Event()
                 download_futures = []
-                for file_info in self._iter_files(dataset.storage):
-                    assert file_info.size is not None
-                    if total_bytes_to_download is None:
-                        progress.update(bytes_task, total=total_downloaded + file_info.size + 1)
-                    target_path = target / Path(file_info.path)
-                    if not force and target_path.exists():
-                        raise FileExistsError(file_info.path)
-                    future = executor.submit(
-                        self._download_file,
-                        dataset.storage,
-                        file_info,
-                        target_path,
-                        progress,
-                        bytes_task,
-                        validate_checksum,
-                    )
-                    download_futures.append(future)
+                try:
+                    for file_info in dataset_info.page.data:
+                        target_path = target / Path(file_info.path)
+                        if not force and target_path.exists():
+                            raise FileExistsError(file_info.path)
+                        future = executor.submit(
+                            self._download_file,
+                            dataset,
+                            file_info,
+                            target_path,
+                            progress=progress if not multiprocessing else None,
+                            task_id=bytes_task if not multiprocessing else None,
+                            validate_checksum=validate_checksum,
+                            chunk_size=chunk_size,
+                        )
+                        download_futures.append(future)
 
-                for future in concurrent.futures.as_completed(download_futures):
-                    total_downloaded += future.result()
+                    for future in concurrent.futures.as_completed(download_futures):
+                        total_downloaded += future.result()
+                        progress.update(bytes_task, completed=total_downloaded)
+                except KeyboardInterrupt:
+                    self.logger.warning("Received KeyboardInterrupt, canceling download workers...")
+                    is_canceled.set()  # type: ignore
+                    for future in download_futures:
+                        future.cancel()
+                    executor.shutdown(wait=True)
+                    raise
 
-            if total_bytes_to_download is None:
-                progress.update(bytes_task, total=total_downloaded, completed=total_downloaded)
+            progress.update(bytes_task, total=total_downloaded, completed=total_downloaded)
 
     def stream_file(
         self,
@@ -287,6 +324,7 @@ class DatasetClient(ServiceClient):
         length: int = -1,
         quiet: bool = False,
         validate_checksum: bool = True,
+        chunk_size: Optional[int] = None,
     ) -> Generator[bytes, None, None]:
         """
         Stream download the contents of a single file from a dataset.
@@ -302,6 +340,8 @@ class DatasetClient(ServiceClient):
         :param length: Number of bytes to read.
         :param quiet: If ``True``, progress won't be displayed.
         :param validate_checksum: If ``True``, the checksum of the downloaded bytes will be verified.
+        :param chunk_size: The size of the buffer (in bytes) to use while downloading each file.
+            Defaults to :data:`DOWNLOAD_CHUNK_SIZE`.
 
         :raises DatasetNotFound: If the dataset can't be found.
         :raises DatasetReadError: If the :data:`~beaker.data_model.dataset.Dataset.storage` hasn't been set.
@@ -318,11 +358,8 @@ class DatasetClient(ServiceClient):
         >>> with open(tmp_path / "squad-train.arrow", "wb") as f:
         ...     for chunk in beaker.dataset.stream_file(squad_dataset_name, "squad-train.arrow", quiet=True):
         ...         total_bytes += f.write(chunk)
-        <BLANKLINE>
         """
         dataset = self.resolve_dataset(dataset, ensure_storage=True)
-        assert dataset.storage is not None
-
         file_info = file if isinstance(file, FileInfo) else self.file_info(dataset, file)
 
         from ..progress import get_unsized_dataset_fetch_progress
@@ -330,11 +367,12 @@ class DatasetClient(ServiceClient):
         with get_unsized_dataset_fetch_progress(quiet=quiet) as progress:
             task_id = progress.add_task("Downloading", total=None)
             for bytes_chunk in self._stream_file(
-                dataset.storage,
+                dataset,
                 file_info,
                 offset=offset,
                 length=length,
                 validate_checksum=validate_checksum,
+                chunk_size=chunk_size,
             ):
                 progress.update(task_id, advance=len(bytes_chunk))
                 yield bytes_chunk
@@ -347,6 +385,7 @@ class DatasetClient(ServiceClient):
         length: int = -1,
         quiet: bool = False,
         validate_checksum: bool = True,
+        chunk_size: Optional[int] = None,
     ) -> bytes:
         """
         Download the contents of a single file from a dataset.
@@ -361,6 +400,8 @@ class DatasetClient(ServiceClient):
         :param length: Number of bytes to read.
         :param quiet: If ``True``, progress won't be displayed.
         :param validate_checksum: If ``True``, the checksum of the downloaded bytes will be verified.
+        :param chunk_size: The size of the buffer (in bytes) to use while downloading each file.
+            Defaults to :data:`DOWNLOAD_CHUNK_SIZE`.
 
         :raises DatasetNotFound: If the dataset can't be found.
         :raises DatasetReadError: If the :data:`~beaker.data_model.dataset.Dataset.storage` hasn't been set.
@@ -374,7 +415,6 @@ class DatasetClient(ServiceClient):
         :examples:
 
         >>> contents = beaker.dataset.get_file(squad_dataset_name, "squad-train.arrow", quiet=True)
-        <BLANKLINE>
         """
 
         @retriable(recoverable_errors=(RequestException, ChecksumFailedError))
@@ -387,6 +427,7 @@ class DatasetClient(ServiceClient):
                     length=length,
                     quiet=quiet,
                     validate_checksum=validate_checksum,
+                    chunk_size=chunk_size,
                 )
             )
 
@@ -408,23 +449,44 @@ class DatasetClient(ServiceClient):
         """
         dataset = self.resolve_dataset(dataset, ensure_storage=True)
         assert dataset.storage is not None
-        response = self.request(
-            f"datasets/{dataset.storage.id}/files/{file_name}",
-            method="HEAD",
-            token=dataset.storage.token,
-            base_url=dataset.storage.address,
-            exceptions_for_status={404: FileNotFoundError(file_name)},
-        )
-        size_str = response.headers.get(self.HEADER_CONTENT_LENGTH)
-        size = int(size_str) if size_str else None
-        return FileInfo(
-            path=file_name,
-            digest=Digest(response.headers[self.HEADER_DIGEST]),
-            updated=datetime.strptime(
-                response.headers[self.HEADER_LAST_MODIFIED], "%a, %d %b %Y %H:%M:%S %Z"
-            ),
-            size=size,
-        )
+        if dataset.storage.scheme == "fh":
+            response = self.request(
+                f"datasets/{dataset.storage.id}/files/{file_name}",
+                method="HEAD",
+                token=dataset.storage.token,
+                base_url=dataset.storage.base_url,
+                exceptions_for_status={404: FileNotFoundError(file_name)},
+            )
+            size_str = response.headers.get(self.HEADER_CONTENT_LENGTH)
+            size = int(size_str) if size_str else None
+            return FileInfo(
+                path=file_name,
+                digest=Digest.from_encoded(response.headers[self.HEADER_DIGEST]),
+                updated=datetime.strptime(
+                    response.headers[self.HEADER_LAST_MODIFIED], "%a, %d %b %Y %H:%M:%S %Z"
+                ),
+                size=size,
+            )
+        else:
+            # TODO (epwalsh): make a HEAD request once Beaker supports that
+            # (https://github.com/allenai/beaker/issues/2961)
+            response = self.request(
+                f"datasets/{dataset.id}/files/{urllib.parse.quote(file_name, safe='')}",
+                stream=True,
+                exceptions_for_status={404: FileNotFoundError(file_name)},
+            )
+            response.close()
+            size_str = response.headers.get(self.HEADER_CONTENT_LENGTH)
+            size = int(size_str) if size_str else None
+            digest = response.headers.get(self.HEADER_DIGEST)
+            return FileInfo(
+                path=file_name,
+                digest=None if digest is None else Digest.from_encoded(digest),
+                updated=datetime.strptime(
+                    response.headers[self.HEADER_LAST_MODIFIED], "%a, %d %b %Y %H:%M:%S %Z"
+                ),
+                size=size,
+            )
 
     def delete(self, dataset: Union[str, Dataset]):
         """
@@ -535,7 +597,7 @@ class DatasetClient(ServiceClient):
                         target_path,
                         progress,
                         bytes_task,
-                        True,
+                        ignore_errors=True,
                     )
                     future_to_path[future] = path
 
@@ -583,11 +645,12 @@ class DatasetClient(ServiceClient):
                 progress.update(task_id, total=size)
             self._upload_file(dataset, size, source, target, progress, task_id)
 
-    def ls(self, dataset: Union[str, Dataset]) -> Generator[FileInfo, None, None]:
+    def ls(self, dataset: Union[str, Dataset], prefix: Optional[str] = None) -> List[FileInfo]:
         """
         List files in a dataset.
 
         :param dataset: The dataset ID, name, or object.
+        :param prefix: An optional path prefix to filter by.
 
         :raises DatasetNotFound: If the dataset can't be found.
         :raises DatasetReadError: If the :data:`~beaker.data_model.dataset.Dataset.storage` hasn't been set.
@@ -596,13 +659,15 @@ class DatasetClient(ServiceClient):
             Beaker server.
         """
         dataset = self.resolve_dataset(dataset)
-        if dataset.storage is None:
-            # Might need to get dataset again if 'storage' hasn't been set yet.
-            dataset = self.get(dataset.id)
-        if dataset.storage is None:
-            raise DatasetReadError(dataset.id)
-        for file_info in self._iter_files(dataset.storage):
-            yield file_info
+        query = {} if prefix is None else {"prefix": prefix}
+        info = DatasetInfo.from_json(
+            self.request(
+                f"datasets/{dataset.id}/files",
+                query=query,
+                exceptions_for_status={404: DatasetNotFound(self._not_found_err_msg(dataset.id))},
+            ).json()
+        )
+        return list(info.page.data)
 
     def size(self, dataset: Union[str, Dataset]) -> int:
         """
@@ -615,11 +680,14 @@ class DatasetClient(ServiceClient):
         :raises RequestException: Any other exception that can occur when contacting the
             Beaker server.
         """
-        total = 0
-        for file_info in self.ls(dataset):
-            assert file_info.size is not None
-            total += file_info.size
-        return total
+        dataset = self.resolve_dataset(dataset)
+        info = DatasetInfo.from_json(
+            self.request(
+                f"datasets/{dataset.id}/files",
+                exceptions_for_status={404: DatasetNotFound(self._not_found_err_msg(dataset.id))},
+            ).json()
+        )
+        return info.size.bytes
 
     def rename(self, dataset: Union[str, Dataset], name: str) -> Dataset:
         """
@@ -680,6 +748,10 @@ class DatasetClient(ServiceClient):
         from ..progress import BufferedReaderWithProgress
 
         assert dataset.storage is not None
+        if dataset.storage.scheme != "fh":
+            raise NotImplementedError(
+                f"Datasets API is not implemented for '{dataset.storage.scheme}' backend yet"
+            )
 
         source_file_wrapper: BufferedReaderWithProgress
         if isinstance(source, (str, Path, os.PathLike)):
@@ -705,7 +777,7 @@ class DatasetClient(ServiceClient):
                         "uploads",
                         method="POST",
                         token=dataset.storage.token,
-                        base_url=dataset.storage.address,
+                        base_url=dataset.storage.base_url,
                     )
                     return response.headers[self.HEADER_UPLOAD_ID]
 
@@ -725,7 +797,7 @@ class DatasetClient(ServiceClient):
                             method="PATCH",
                             data=chunk,
                             token=dataset.storage.token,
-                            base_url=dataset.storage.address,
+                            base_url=dataset.storage.base_url,
                             headers={
                                 self.HEADER_UPLOAD_LENGTH: str(size),
                                 self.HEADER_UPLOAD_OFFSET: str(written),
@@ -752,7 +824,7 @@ class DatasetClient(ServiceClient):
                     method="PUT",
                     data=body,
                     token=dataset.storage.token,
-                    base_url=dataset.storage.address,
+                    base_url=dataset.storage.base_url,
                     headers=None if not digest else {self.HEADER_DIGEST: digest},
                     stream=body is not None,
                     exceptions_for_status={
@@ -767,41 +839,15 @@ class DatasetClient(ServiceClient):
         finally:
             source_file_wrapper.close()
 
-    def _iter_files(self, storage: DatasetStorage) -> Generator[FileInfo, None, None]:
-        from collections import deque
-
-        files: Deque[FileInfo] = deque([])
-        last_request: bool = False
-        cursor: Optional[str] = ""
-        while files or not last_request:
-            if files:
-                yield files.popleft()
-            else:
-                manifest = DatasetManifest.from_json(
-                    self.request(
-                        f"datasets/{storage.id}/manifest",
-                        method="GET",
-                        token=storage.token,
-                        base_url=storage.address,
-                        query={"cursor": cursor, "path": "", "url": True},
-                    ).json()
-                )
-                files.extend(manifest.files)
-                cursor = manifest.cursor
-                if not cursor:
-                    last_request = True
-
     def _stream_file(
         self,
-        storage: DatasetStorage,
-        file_info: FileInfo,
-        chunk_size: int = 1024,
+        dataset: Dataset,
+        file: FileInfo,
+        chunk_size: Optional[int] = None,
         offset: int = 0,
         length: int = -1,
         validate_checksum: bool = True,
     ) -> Generator[bytes, None, None]:
-        import hashlib
-
         def stream_file() -> Generator[bytes, None, None]:
             headers = {}
             if offset > 0 and length > 0:
@@ -809,25 +855,31 @@ class DatasetClient(ServiceClient):
             elif offset > 0:
                 headers["Range"] = f"bytes={offset}-"
             response = self.request(
-                f"datasets/{storage.id}/files/{file_info.path}",
+                f"datasets/{dataset.id}/files/{urllib.parse.quote(file.path, safe='')}",
                 method="GET",
                 stream=True,
                 headers=headers,
-                token=storage.token,
-                base_url=storage.address,
-                exceptions_for_status={404: FileNotFoundError(file_info.path)},
+                exceptions_for_status={404: FileNotFoundError(file.path)},
             )
-            for chunk in response.iter_content(chunk_size=chunk_size):
+            for chunk in response.iter_content(chunk_size=chunk_size or self.DOWNLOAD_CHUNK_SIZE):
                 yield chunk
 
-        sha256_hash = hashlib.sha256() if (offset == 0 and validate_checksum) else None
+        if is_canceled is not None and is_canceled.is_set():  # type: ignore
+            raise ThreadCanceledError
+
+        contents_hash = None
+        if offset == 0 and validate_checksum and file.digest is not None:
+            contents_hash = file.digest.new_hasher()
+
         retries = 0
         while True:
             try:
                 for chunk in stream_file():
+                    if is_canceled is not None and is_canceled.is_set():  # type: ignore
+                        raise ThreadCanceledError
                     offset += len(chunk)
-                    if sha256_hash is not None:
-                        sha256_hash.update(chunk)
+                    if contents_hash is not None:
+                        contents_hash.update(chunk)
                     yield chunk
                 break
             except RequestException as err:
@@ -838,22 +890,25 @@ class DatasetClient(ServiceClient):
                     raise
 
         # Validate digest.
-        if sha256_hash is not None:
-            digest = Digest(sha256_hash.digest())
-            if file_info.digest != digest:
+        if file.digest is not None and contents_hash is not None:
+            actual_digest = Digest.from_decoded(
+                contents_hash.digest(), algorithm=file.digest.algorithm
+            )
+            if actual_digest != file.digest:
                 raise ChecksumFailedError(
-                    f"Checksum for '{file_info.path}' failed. "
-                    f"Expected '{file_info.digest}', got '{digest}'."
+                    f"Checksum for '{file.path}' failed. "
+                    f"Expected '{file.digest}', got '{actual_digest}'."
                 )
 
     def _download_file(
         self,
-        storage: DatasetStorage,
-        file_info: FileInfo,
+        dataset: Dataset,
+        file: FileInfo,
         target_path: Path,
-        progress: "Progress",
-        task_id: "TaskID",
+        progress: Optional["Progress"] = None,
+        task_id: Optional["TaskID"] = None,
         validate_checksum: bool = True,
+        chunk_size: Optional[int] = None,
     ) -> int:
         import tempfile
 
@@ -861,8 +916,12 @@ class DatasetClient(ServiceClient):
         target_dir = target_path.parent
         target_dir.mkdir(exist_ok=True, parents=True)
 
+        def on_failure():
+            if progress is not None and task_id is not None:
+                progress.advance(task_id, -total_bytes)
+
         @retriable(
-            on_failure=lambda: progress.advance(task_id, -total_bytes),
+            on_failure=on_failure,
             recoverable_errors=(RequestException, ChecksumFailedError),
         )
         def download() -> int:
@@ -873,11 +932,15 @@ class DatasetClient(ServiceClient):
             )
             try:
                 for chunk in self._stream_file(
-                    storage, file_info, validate_checksum=validate_checksum
+                    dataset,
+                    file,
+                    validate_checksum=validate_checksum,
+                    chunk_size=chunk_size,
                 ):
                     total_bytes += len(chunk)
                     tmp_target.write(chunk)
-                    progress.update(task_id, advance=len(chunk))
+                    if progress is not None and task_id is not None:
+                        progress.update(task_id, advance=len(chunk))
                 os.replace(tmp_target.name, target_path)
             finally:
                 tmp_target.close()
