@@ -1,11 +1,14 @@
 from collections import defaultdict
 from datetime import datetime
-from typing import Dict, Generator, List, Optional, Union
+from typing import Any, Dict, Generator, List, Optional, Type, TypeVar, Union
 
 from ..data_model import *
 from ..exceptions import *
+from ..util import format_cursor
 from .service_client import ServiceClient
 
+T = TypeVar("T")
+
 
 class WorkspaceClient(ServiceClient):
     """
@@ -52,7 +55,7 @@ class WorkspaceClient(ServiceClient):
             raise
 
     def create(
-        self, workspace: str, description: Optional[str] = None, public: bool = False
+        self, workspace: str, *, description: Optional[str] = None, public: bool = False
     ) -> Workspace:
         """
         Create a workspace.
@@ -210,13 +213,84 @@ class WorkspaceClient(ServiceClient):
             },
         )
 
-    def list(
+    def _paginated_requests(
+        self,
+        page_class: Type[BasePage[T]],
+        path: str,
+        query: Dict[str, Any],
+        limit: Optional[int] = None,
+        workspace_name: Optional[str] = None,
+    ) -> Generator[T, None, None]:
+        if limit:
+            query["limit"] = str(limit)
+
+        exceptions_for_status: Optional[Dict[int, Exception]] = (
+            None
+            if workspace_name is None
+            else {404: WorkspaceNotFound(self._not_found_err_msg(workspace_name))}
+        )
+
+        count = 0
+        while True:
+            page = page_class.from_json(
+                self.request(
+                    path,
+                    method="GET",
+                    query=query,
+                    exceptions_for_status=exceptions_for_status,
+                ).json()
+            )
+            for x in page.data:
+                count += 1
+                yield x
+                if limit is not None and count >= limit:
+                    return
+
+            query["cursor"] = page.next_cursor or page.next
+            if not query["cursor"]:
+                break
+
+    def iter(
         self,
         org: Optional[Union[str, Organization]] = None,
+        *,
         author: Optional[Union[str, Account]] = None,
         match: Optional[str] = None,
         archived: Optional[bool] = None,
         limit: Optional[int] = None,
+        sort_by: WorkspaceSort = WorkspaceSort.created,
+        descending: bool = True,
+        cursor: int = 0,
+    ) -> Generator[Workspace, None, None]:
+        org = self.resolve_org(org)
+        query: Dict[str, str] = {
+            "org": org.id,
+            "field": str(sort_by),
+            "order": "descending" if descending else "ascending",
+            "cursor": format_cursor(cursor),
+        }
+        if author is not None:
+            query["author"] = (
+                author.name if isinstance(author, Account) else self.beaker.account.get(author).name
+            )
+        if match is not None:
+            query["q"] = match
+        if archived is not None:
+            query["archived"] = str(archived).lower()
+
+        yield from self._paginated_requests(WorkspacePage, "workspaces", query, limit=limit)
+
+    def list(
+        self,
+        org: Optional[Union[str, Organization]] = None,
+        *,
+        author: Optional[Union[str, Account]] = None,
+        match: Optional[str] = None,
+        archived: Optional[bool] = None,
+        limit: Optional[int] = None,
+        sort_by: WorkspaceSort = WorkspaceSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> List[Workspace]:
         """
         List workspaces belonging to an organization.
@@ -227,6 +301,9 @@ class WorkspaceClient(ServiceClient):
         :param match: Only include workspaces matching the text.
         :param archived: Only include/exclude archived workspaces.
         :param limit: Limit the number of workspaces returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises OrganizationNotFound: If the organization doesn't exist.
         :raises OrganizationNotSet: If neither ``org`` nor
@@ -236,45 +313,28 @@ class WorkspaceClient(ServiceClient):
         :raises RequestException: Any other exception that can occur when contacting the
             Beaker server.
         """
-        org = self.resolve_org(org)
-        workspaces: List[Workspace] = []
-        cursor: Optional[str] = None
-        query: Dict[str, str] = {"org": org.id}
-        if author is not None:
-            query["author"] = (
-                author.name if isinstance(author, Account) else self.beaker.account.get(author).name
+        return list(
+            self.iter(
+                org=org,
+                author=author,
+                match=match,
+                archived=archived,
+                limit=limit,
+                sort_by=sort_by,
+                descending=descending,
+                cursor=cursor,
             )
-        if match is not None:
-            query["q"] = match
-        if archived is not None:
-            query["archived"] = str(archived).lower()
-        if limit:
-            query["limit"] = str(limit)
-
-        while True:
-            query["cursor"] = cursor or ""
-            page = WorkspacePage.from_json(
-                self.request(
-                    "workspaces",
-                    method="GET",
-                    query=query,
-                ).json()
-            )
-            workspaces.extend(page.data)
-            cursor = page.next_cursor or page.next
-            if not cursor:
-                break
-            if limit and len(workspaces) >= limit:
-                workspaces = workspaces[:limit]
-                break
-
-        return workspaces
+        )
 
     def iter_images(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         limit: Optional[int] = None,
+        sort_by: ImageSort = ImageSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> Generator[Image, None, None]:
         """
         Iterate over the images in a workspace.
@@ -283,6 +343,9 @@ class WorkspaceClient(ServiceClient):
             :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
         :param match: Only include images matching the text.
         :param limit: Limit the number of images returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -292,40 +355,31 @@ class WorkspaceClient(ServiceClient):
             Beaker server.
         """
         workspace_name = self.resolve_workspace(workspace, read_only_ok=True).full_name
-        cursor: Optional[str] = None
-        query: Dict[str, str] = {}
+        query: Dict[str, str] = {
+            "field": str(sort_by),
+            "order": "descending" if descending else "ascending",
+            "cursor": format_cursor(cursor),
+        }
         if match is not None:
             query["q"] = match
 
-        count = 0
-        while True:
-            query["cursor"] = cursor or ""
-            page = ImagesPage.from_json(
-                self.request(
-                    f"workspaces/{self.url_quote(workspace_name)}/images",
-                    method="GET",
-                    query=query,
-                    exceptions_for_status={
-                        404: WorkspaceNotFound(self._not_found_err_msg(workspace_name))
-                    },
-                ).json()
-            )
-
-            for image in page.data:
-                count += 1
-                yield image
-                if limit is not None and count >= limit:
-                    return
-
-            cursor = page.next_cursor or page.next
-            if not cursor:
-                break
+        yield from self._paginated_requests(
+            ImagesPage,
+            f"workspaces/{self.url_quote(workspace_name)}/images",
+            query,
+            limit=limit,
+            workspace_name=workspace_name,
+        )
 
     def images(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         limit: Optional[int] = None,
+        sort_by: ImageSort = ImageSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> List[Image]:
         """
         List the images in a workspace.
@@ -334,6 +388,9 @@ class WorkspaceClient(ServiceClient):
             :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
         :param match: Only include images matching the text.
         :param limit: Limit the number of images returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -342,13 +399,26 @@ class WorkspaceClient(ServiceClient):
         :raises RequestException: Any other exception that can occur when contacting the
             Beaker server.
         """
-        return list(self.iter_images(workspace=workspace, match=match, limit=limit))
+        return list(
+            self.iter_images(
+                workspace=workspace,
+                match=match,
+                limit=limit,
+                sort_by=sort_by,
+                descending=descending,
+                cursor=cursor,
+            )
+        )
 
     def iter_experiments(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         limit: Optional[int] = None,
+        sort_by: ExperimentSort = ExperimentSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> Generator[Experiment, None, None]:
         """
         Iterate over the experiments in a workspace.
@@ -357,6 +427,9 @@ class WorkspaceClient(ServiceClient):
             :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
         :param match: Only include experiments matching the text.
         :param limit: Limit the number of experiments returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -366,40 +439,31 @@ class WorkspaceClient(ServiceClient):
             Beaker server.
         """
         workspace_name = self.resolve_workspace(workspace, read_only_ok=True).full_name
-        cursor: Optional[str] = None
-        query: Dict[str, str] = {}
+        query: Dict[str, str] = {
+            "field": str(sort_by),
+            "order": "descending" if descending else "ascending",
+            "cursor": format_cursor(cursor),
+        }
         if match is not None:
             query["q"] = match
 
-        count = 0
-        while True:
-            query["cursor"] = cursor or ""
-            page = ExperimentsPage.from_json(
-                self.request(
-                    f"workspaces/{self.url_quote(workspace_name)}/experiments",
-                    method="GET",
-                    query=query,
-                    exceptions_for_status={
-                        404: WorkspaceNotFound(self._not_found_err_msg(workspace_name))
-                    },
-                ).json()
-            )
-
-            for experiment in page.data:
-                count += 1
-                yield experiment
-                if limit is not None and count >= limit:
-                    return
-
-            cursor = page.next_cursor or page.next
-            if not cursor:
-                break
+        yield from self._paginated_requests(
+            ExperimentsPage,
+            f"workspaces/{self.url_quote(workspace_name)}/experiments",
+            query,
+            limit=limit,
+            workspace_name=workspace_name,
+        )
 
     def experiments(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         limit: Optional[int] = None,
+        sort_by: ExperimentSort = ExperimentSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> List[Experiment]:
         """
         List the experiments in a workspace.
@@ -408,6 +472,9 @@ class WorkspaceClient(ServiceClient):
             :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
         :param match: Only include experiments matching the text.
         :param limit: Limit the number of experiments returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -416,15 +483,28 @@ class WorkspaceClient(ServiceClient):
         :raises RequestException: Any other exception that can occur when contacting the
             Beaker server.
         """
-        return list(self.iter_experiments(workspace=workspace, match=match, limit=limit))
+        return list(
+            self.iter_experiments(
+                workspace=workspace,
+                match=match,
+                limit=limit,
+                sort_by=sort_by,
+                descending=descending,
+                cursor=cursor,
+            )
+        )
 
     def iter_datasets(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         results: Optional[bool] = None,
         uncommitted: Optional[bool] = None,
         limit: Optional[int] = None,
+        sort_by: DatasetSort = DatasetSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> Generator[Dataset, None, None]:
         """
         Iterate over the datasets in a workspace.
@@ -435,6 +515,9 @@ class WorkspaceClient(ServiceClient):
         :param results: Only include/exclude experiment result datasets.
         :param uncommitted: Only include/exclude uncommitted datasets.
         :param limit: Limit the number of datasets returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -444,8 +527,11 @@ class WorkspaceClient(ServiceClient):
             Beaker server.
         """
         workspace_name = self.resolve_workspace(workspace, read_only_ok=True).full_name
-        cursor: Optional[str] = None
-        query: Dict[str, str] = {}
+        query: Dict[str, str] = {
+            "field": str(sort_by),
+            "order": "descending" if descending else "ascending",
+            "cursor": format_cursor(cursor),
+        }
         if match is not None:
             query["q"] = match
         if results is not None:
@@ -453,37 +539,25 @@ class WorkspaceClient(ServiceClient):
         if uncommitted is not None:
             query["committed"] = str(not uncommitted).lower()
 
-        count = 0
-        while True:
-            query["cursor"] = cursor or ""
-            page = DatasetsPage.from_json(
-                self.request(
-                    f"workspaces/{self.url_quote(workspace_name)}/datasets",
-                    method="GET",
-                    query=query,
-                    exceptions_for_status={
-                        404: WorkspaceNotFound(self._not_found_err_msg(workspace_name))
-                    },
-                ).json()
-            )
-
-            for dataset in page.data:
-                count += 1
-                yield dataset
-                if limit is not None and count >= limit:
-                    return
-
-            cursor = page.next_cursor or page.next
-            if not cursor:
-                break
+        yield from self._paginated_requests(
+            DatasetsPage,
+            f"workspaces/{self.url_quote(workspace_name)}/datasets",
+            query,
+            limit=limit,
+            workspace_name=workspace_name,
+        )
 
     def datasets(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         match: Optional[str] = None,
         results: Optional[bool] = None,
         uncommitted: Optional[bool] = None,
         limit: Optional[int] = None,
+        sort_by: DatasetSort = DatasetSort.created,
+        descending: bool = True,
+        cursor: int = 0,
     ) -> List[Dataset]:
         """
         List the datasets in a workspace.
@@ -494,6 +568,9 @@ class WorkspaceClient(ServiceClient):
         :param results: Only include/exclude experiment result datasets.
         :param uncommitted: Only include/exclude uncommitted datasets.
         :param limit: Limit the number of datasets returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -509,6 +586,9 @@ class WorkspaceClient(ServiceClient):
                 results=results,
                 uncommitted=uncommitted,
                 limit=limit,
+                sort_by=sort_by,
+                descending=descending,
+                cursor=cursor,
             )
         )
 
@@ -538,12 +618,26 @@ class WorkspaceClient(ServiceClient):
             ).json()["data"]
         ]
 
-    def groups(self, workspace: Optional[Union[str, Workspace]] = None) -> List[Group]:
+    def iter_groups(
+        self,
+        workspace: Optional[Union[str, Workspace]] = None,
+        *,
+        match: Optional[str] = None,
+        limit: Optional[int] = None,
+        sort_by: GroupSort = GroupSort.created,
+        descending: bool = True,
+        cursor: int = 0,
+    ) -> Generator[Group, None, None]:
         """
-        List groups in a workspace.
+        Iterate over groups in a workspace.
 
         :param workspace: The Beaker workspace name, or object. If not specified,
             :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
+        :param match: Only include groups matching the text.
+        :param limit: Limit the number of groups returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
 
         :raises WorkspaceNotFound: If the workspace doesn't exist.
         :raises WorkspaceNotSet: If neither ``workspace`` nor
@@ -553,16 +647,60 @@ class WorkspaceClient(ServiceClient):
             Beaker server.
         """
         workspace_name = self.resolve_workspace(workspace, read_only_ok=True).full_name
-        return [
-            Group.from_json(d)
-            for d in self.request(
-                f"workspaces/{self.url_quote(workspace_name)}/groups",
-                method="GET",
-                exceptions_for_status={
-                    404: WorkspaceNotFound(self._not_found_err_msg(workspace_name))
-                },
-            ).json()["data"]
-        ]
+        query: Dict[str, str] = {
+            "field": str(sort_by),
+            "order": "descending" if descending else "ascending",
+            "cursor": format_cursor(cursor),
+        }
+        if match is not None:
+            query["q"] = match
+
+        yield from self._paginated_requests(
+            GroupsPage,
+            f"workspaces/{self.url_quote(workspace_name)}/groups",
+            query,
+            limit=limit,
+            workspace_name=workspace_name,
+        )
+
+    def groups(
+        self,
+        workspace: Optional[Union[str, Workspace]] = None,
+        *,
+        match: Optional[str] = None,
+        limit: Optional[int] = None,
+        sort_by: GroupSort = GroupSort.created,
+        descending: bool = True,
+        cursor: int = 0,
+    ) -> List[Group]:
+        """
+        List groups in a workspace.
+
+        :param workspace: The Beaker workspace name, or object. If not specified,
+            :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` is used.
+        :param match: Only include groups matching the text.
+        :param limit: Limit the number of groups returned.
+        :param sort_by: The field to sort the results by.
+        :param descending: Order the results in descending order according to the ``sort_by`` field.
+        :param cursor: Set the starting cursor for the query. You can use this to paginate the results.
+
+        :raises WorkspaceNotFound: If the workspace doesn't exist.
+        :raises WorkspaceNotSet: If neither ``workspace`` nor
+            :data:`Beaker.config.default_workspace <beaker.Config.default_workspace>` are set.
+        :raises BeakerError: Any other :class:`~beaker.exceptions.BeakerError` type that can occur.
+        :raises RequestException: Any other exception that can occur when contacting the
+            Beaker server.
+        """
+        return list(
+            self.iter_groups(
+                workspace=workspace,
+                match=match,
+                limit=limit,
+                sort_by=sort_by,
+                descending=descending,
+                cursor=cursor,
+            )
+        )
 
     def get_permissions(
         self, workspace: Optional[Union[str, Workspace]] = None
@@ -706,6 +844,7 @@ class WorkspaceClient(ServiceClient):
     def clear(
         self,
         workspace: Optional[Union[str, Workspace]] = None,
+        *,
         groups: bool = True,
         experiments: bool = True,
         images: bool = True,
@@ -733,6 +872,7 @@ class WorkspaceClient(ServiceClient):
             Beaker server.
         """
         import concurrent.futures
+        from itertools import chain
 
         def should_delete(created: Optional[datetime]) -> bool:
             if older_than is None or created is None:
@@ -770,7 +910,11 @@ class WorkspaceClient(ServiceClient):
 
             if datasets:
                 for dataset in filter(
-                    lambda x: should_delete(x.created), self.iter_datasets(workspace)
+                    lambda x: should_delete(x.created),
+                    chain(
+                        self.iter_datasets(workspace),
+                        self.iter_datasets(workspace, uncommitted=True),
+                    ),
                 ):
                     future = executor.submit(self.beaker.dataset.delete, dataset)
                     deletion_futures.append(future)
